/** @file ccp_model.h
 *
 * A model of the simple write-invalidate cache coherence protocol (CCP) used to
 * illustrate the C++TESK facilities on cache coherence (CC) verification using
 * the graph exploration (state enumeration) approach. The idea of the example
 * was borrowed from F. Pong and M. Dubois' paper "Verification Techniques for
 * Cache Coherence Protocols" (ACM Computing Surveys (CSUR), Volume 29 Issue 1,
 * March 1997).
 *
 * @author Alexander Kamkin (kamkin@ispras.ru)
 */

#pragma once

#include <hw/model.hpp>

#include <ccp_coverage.h>

#include <sstream>
#include <vector>

namespace cpptesk {
namespace examples {
namespace ccp {

////////////////////////////////////////////////////////////////////////////////
// Cache
////////////////////////////////////////////////////////////////////////////////

/**
 * A single-block cache model.
 */
class Cache
{
public:
    /// Cache states.
    enum state_t
    {
        /// Invalid (I): invalid copy.
        INVALID,
        /// Shared (S): clean copy shared with other caches.
        SHARED,
        /// Dirty (D): modified and only cached copy.
        DIRTY
    };

    /// Cache operations.
    enum operation_t
    {
        /// Read (R).
        READ,
        /// Write (W).
        WRITE,
        /// Replacement (Z).
        REPLACE
    };

    /// Cache requesters.
    enum requester_t
    {
        /// Request from this cache.
        THIS,
        /// Request from an other cache.
        OTHER
    };

    Cache(): _state(INVALID) {}
    virtual ~Cache() {}

    /// Returns the cache state.
    state_t state() const { return _state; }
    /// Resets the cache state.
    void reset() { _state = INVALID; }

    /// Processes an operation (updates the cache state).
    void process(requester_t r, operation_t o);

protected:
    /// Cache state.
    state_t _state;
};

////////////////////////////////////////////////////////////////////////////////
// Processor
////////////////////////////////////////////////////////////////////////////////

/**
 * A processor model.
 */
class Processor
{
public:
    Processor(unsigned id): _id(id) {}
    virtual ~Processor() {}

    /// Processes an operation.
    virtual void process(unsigned id, Cache::operation_t o)
        { _cache.process((id == _id ? Cache::THIS : Cache::OTHER), o); }

    /// Returns a processor state.
    virtual std::string state() const
    {
        switch(_cache.state())
        {
        case Cache::INVALID: { return "I"; }
        case Cache::SHARED:  { return "S"; }
        case Cache::DIRTY:   { return "D"; }
        }
        return "-";
    }

    /// Reads data.
    void read   (unsigned id) { process(id, Cache::READ);    }
    /// Writes data.
    void write  (unsigned id) { process(id, Cache::WRITE);   }
    /// Replaces data.
    void replace(unsigned id) { process(id, Cache::REPLACE); }

    /// Replace precondition.
    virtual bool pre_replace(unsigned id) const
    {
        if(_cache.state() == Cache::INVALID && id == _id) { return false; }
	if(_cache.state() == Cache::DIRTY   && id != _id) { return false; }
	return true;
    }

    /// Resets the processor state.
    virtual void reset() { _cache.reset(); }

    /// Returns the processor id.
    unsigned id() const { return _id; }

protected:
    Processor(): _id(-1u) {}

    Cache _cache;
    unsigned _id;
};

////////////////////////////////////////////////////////////////////////////////
// Symmetric Multi-Processor
////////////////////////////////////////////////////////////////////////////////

/**
 * A multi-processor model.
 */
class MultiProcessor: public Processor
{
public:
    MultiProcessor(unsigned n)
    {
        for(unsigned i = 0; i < n; i++)
        {
	    Processor processor(i);
            _processors.push_back(processor);
        }
    }

    virtual ~MultiProcessor() {}

    /// Processes an operation.
    virtual void process(unsigned id, Cache::operation_t o)
    {
        for(std::vector<Processor>::iterator i = _processors.begin(); i != _processors.end(); i++)
            { i->process(id, o); }
    }

    /// Returns a composite state.
    virtual std::string state() const
    {
        std::stringstream buffer;

        buffer << "(";
        for(std::vector<Processor>::const_iterator i = _processors.begin(); i != _processors.end(); i++)
            { buffer << (i == _processors.begin() ? "" : ",") << i->state(); }
        buffer << ")";

        return buffer.str();
    }

    /// Replace precondition.
    virtual bool pre_replace(unsigned id) const
    {
        return _processors[id].pre_replace(id);
    }

    /// Resets the multi-processor state.
    virtual void reset()
    {
        for(std::vector<Processor>::iterator i = _processors.begin(); i != _processors.end(); i++)
	    { i->reset(); }
    }

    /// Returns the number of processors.
    size_t size() const { return _processors.size(); }

    const std::vector<Processor>& processors() const { return _processors; }

protected:
    std::vector<Processor> _processors;
};

////////////////////////////////////////////////////////////////////////////////
// Cache Coherence Model
////////////////////////////////////////////////////////////////////////////////

class CacheCoherenceModel: public MultiProcessor
{
public:
    CacheCoherenceModel(unsigned n): MultiProcessor(n), _coverage(*this) {}
    virtual ~CacheCoherenceModel() {}

    virtual void initialize() {}
    virtual void finalize()   {}

    /// Data integrity condition.
    virtual bool invariant()
    {
        // TODO: D and S are mutually exclusive
        // TODO: D and D are mutually exclusive
        return true;
    }

    /// Returns the true state.
    std::string getState() const;
    /// Exploits the sequence-to-multiset abstraction.
    std::string getMultiSetState() const;
    /// Exploits the sequence-to-set abstraction.
    std::string getSetState() const;

    void trackCoverage(void) { _coverage.track_cov(); }

protected:
    std::vector<std::string> getStateVector() const;
    std::string toString(const std::vector<std::string> &state) const;

    CacheCoherenceCoverage _coverage;
};

}}} // namespace cpptesk::examples::ccp
