/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <dirent.h>
#include <fcntl.h>

#ifdef unix
  #include <unistd.h>
  #include <sys/wait.h>
#endif
#include <time.h>

#include <sys/types.h>

#include <errno.h>

#include <atl/integer.h>
#include <atl/string.h>
#include <atl/stringbuffer.h>
#include <atl/list.h>
#include <atl/arraylist.h>
#include <atl/set.h>

#include <utils/assertion.h>
#include <utils/network.h>
#include <ts/system.h>

#include "command.h"
#include "mcc.h"

/******************************************************************************/
/**                           Internal Parameters                            **/
/******************************************************************************/
#define DEFAULT_NUMBER_OF_ATTEMPTS 10
#define DEFAULT_TIMEOUT_BETWEEN_ATTEMPTS 3500

#define BUFF_SIZE 1024
#define LINE_SIZE 255
#define ADDRESS_SIZE 255
#define PORT_SIZE 10
#define NUMBER_SIZE 10

#define NODE_ALLOC_BY 50
#define TESTRUN_ALLOC_BY 20

#define DEFAULT_THREAD_SLEEPTIME 1000
#define DEFAULT_TURNS_LEFT_FOR_INFO 20

#define LOG_FILENAME "synchronizer.log"

/******************************************************************************/
/**                              Error Strings                               **/
/******************************************************************************/
/*
 * error strings for setNodeList
 */
#define NULL_COORD_SETNODE_ERROR       "NULL pointer"
#define FILEOPEN_COORD_SETNODE_ERROR   "Can't open node list file"
#define FILECLOSE_COORD_SETNODE_ERROR  "Can't close node list file"
#define FORMAT_COORD_SETNODE_ERROR     "Wrong format"
#define MEMORY_COORD_SETNODE_ERROR     "Out of memory" /* changed to assertion */
/*
 * error strings for setTopology
 */
#define NULL_COORD_SETTOPOLOGY_ERROR         NULL_COORD_SETNODE_ERROR
#define FILEOPEN_COORD_SETTOPOLOGY_ERROR     "Can't open topology file"
#define FILECLOSE_COORD_SETTOPOLOGY_ERROR    "Can't close topology file"
#define FORMAT_COORD_SETTOPOLOGY_ERROR       FORMAT_COORD_SETNODE_ERROR
#define MEMORY_COORD_SETTOPOLOGY_ERROR       MEMORY_COORD_SETNODE_ERROR /* changed to assertion */
#define NONODE_COORD_SETTOPOLOGY_ERROR       "Node list is empty, use setNodeList first"
#define RANGE_COORD_SETTOPOLOGY_ERROR        "Node number %d is out of range (%d node(s) in node list)"
#define NODEEXISTS_COORD_SETTOPOLOGY_ERROR   "Node %d already exists in topology"
#define DUPLICATE_COORD_SETTOPOLOGY_ERROR    "Duplicate %s in the list of out nodes"
/*
 * error strings for saveLog & loadLog
 */
#define NULL_COORD_LOG_ERROR         NULL_COORD_SETNODE_ERROR
#define NONODE_COORD_LOG_ERROR       NONODE_COORD_SETTOPOLOGY_ERROR
#define RANGE_COORD_LOG_ERROR        "Node number %d is out of range (%d node(s) in node list)"
/*
 * error strings for RunScript
 */
#define FORK_RUNSCRIPT_ERROR        "Can not fork"
#define EXEC_RUNSCRIPT_ERROR        "Can not exec"
#define WAITPID_RUNSCRIPT_ERROR     "Can not wait for script execution"
#define NOTNORMALLY_RUNSCRIPT_ERROR "The script has not finished normally"
#define NOTZERO_RUNSCRIPT_ERROR     "Nonzero exit code of the script"
#define NOEXEC_CODE 255
/*
 * error strings for StartTest & StopTest
 */
#define MEMORY_TABLE_COORD_STARTTEST_ERROR     "Can not create Test Run Table: out of memory"
#define MEMORY_TESTRUN_COORD_STARTTEST_ERROR   "Can not add Test Run: out of memory"
#define CONNECT_MCC_ERROR                      "Some nodes could not be connected"
#define START_MCC_ERROR                        "TestRun is not configured"
#define STOP_MCC_ERROR                         "TestRun is not started yet"
#define NO_TESTRUN_MCC_ERROR                   "TestRun is not found"
/*
 * error strings coord_Init
 */
#define GETINFO_THREAD_INIT_ERROR  "Can not start GetInfo thread"

/******************************************************************************/
/**                               NodeStatus                                 **/
/******************************************************************************/
typedef enum NodeStatus {
      NODE_INITIAL
    , NODE_CONFIGURED
    , TEST_STARTED
    , TEST_STOPPED
} NodeStatus;

/******************************************************************************/
String *toString_NodeStatus( NodeStatus status )
{
    switch(status) {
        case NODE_INITIAL:    return create_String( "NODE_INITIAL" );
        case NODE_CONFIGURED: return create_String( "NODE_CONFIGURED" );
        case TEST_STARTED:    return create_String( "TEST_STARTED" );
        case TEST_STOPPED:    return create_String( "TEST_STOPPED" );
        default:              return format_String( "unknown(%d)", status );
    }
}

/******************************************************************************/
static String *to_JSON_String_NodeStatus( NodeStatus status )
{
    String *res;

    switch(status) {
        case NODE_INITIAL:    res = create_String("\"ready\""); break;
        case NODE_CONFIGURED: res = create_String("\"configured\""); break;
        case TEST_STARTED:    res = create_String("\"test started\""); break;
        case TEST_STOPPED:    res = create_String("\"test stopped\""); break;
        default:              res = create_String("\"no information\""); break;
    }

    return res;
}

/******************************************************************************/
/**                                  Node                                    **/
/******************************************************************************/
typedef struct Node {
     String *address;
     unsigned short mccPort;  /* port for communication with coordinator */
     unsigned short inPort;   /* port for input connections for synchronizer */

     bool loadLog;
     bool saveLog;
     String *logFileName;

     NodeStatus status;
     Set *inConections;
     Set *outConections;

     String *info;
     int connection; /* socket if keep connection, -1 otherwise */
} Node;

/******************************************************************************/
static void init_Node( Node *node, String *address, unsigned short mccPort, unsigned short inPort )
{
    assertion( node != NULL, "init_Node: node == NULL" );

    node->address = address;
    node->mccPort = mccPort;
    node->inPort = inPort;

    node->loadLog = false;
    node->saveLog = false;
    node->logFileName = NULL;
    node->status = NODE_INITIAL;
    node->inConections = create_Set( &type_String );
    node->outConections = create_Set( &type_String );
    node->info = NULL;
    node->connection = -1;
}

/******************************************************************************/
static void setLog_Node( Node *node, bool loadLog, bool saveLog, String *logFileName )
{
    assertion( node != NULL, "setLog_Node: node == NULL" );
    node->loadLog = loadLog;
    node->saveLog = saveLog;
    node->logFileName = logFileName;
}

/******************************************************************************/
static void destroy_Node( Node *node )
{
    assertion( node != NULL, "destroy_Node: node == NULL" );

    destroy( node->address );
    destroy( node->logFileName );
    destroy( node->inConections );
    destroy( node->outConections );
    destroy( node->info );
}

static String *toString_Node( Node *node )
{
    StringBuffer *buf = NULL;

    if( node == NULL ) return create_String( "NULL" );
    buf = create_StringBuffer();

    append_StringBuffer( r(buf), "address:" );
    appendString_StringBuffer( r(buf)
                             , node->address != NULL ? r(node->address)
                                                     : create_String("NULL")
                             );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "mccPort:" );
    appendFormat_StringBuffer( r(buf), "%d", node->mccPort );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "inPort:" );
    appendFormat_StringBuffer( r(buf), "%d", node->inPort );

    append_StringBuffer( r(buf), ", " );
    append_StringBuffer( r(buf), "loadLog:" );
    appendFormat_StringBuffer( r(buf), "%s", node->loadLog ? "true" : "false" );

    append_StringBuffer( r(buf), ", " );
    append_StringBuffer( r(buf), "saveLog:" );
    appendFormat_StringBuffer( r(buf), "%s", node->saveLog ? "true" : "false" );

    append_StringBuffer( r(buf), ", " );
    append_StringBuffer( r(buf), "logFileName:" );
    appendString_StringBuffer( r(buf)
                             , node->logFileName != NULL ? r(node->logFileName)
                                                         : create_String("NULL")
                             );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "status:");
    appendString_StringBuffer( r(buf), toString_NodeStatus( node->status ) );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "inConections:");
    appendString_StringBuffer( r(buf), toString_Object( node->inConections ) );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "outConections:");
    appendString_StringBuffer( r(buf), toString_Object( node->outConections ) );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "info:");
    appendString_StringBuffer( r(buf)
                             , node->info != NULL ? r(node->info)
                                                  : create_String("NULL")
                             );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "connection:" );
    appendFormat_StringBuffer( r(buf), "%d", node->connection );

    return toString_Object(buf);
}
/******************************************************************************/
/**                               CommandQueue                               **/
/******************************************************************************/
typedef enum Command {
      NOCMD
    , CONFIG
    , START
    , STOP
} Command;

/******************************************************************************/
static String *toString_Command( Command cmd )
{
    switch(cmd) {
        case NOCMD:  return create_String( "NOCMD" );
        case CONFIG: return create_String( "CONFIG" );
        case START:  return create_String( "START" );
        case STOP:   return create_String( "STOP" );
        defaulf:     return format_String( "unknown(%d)", cmd );
    }
}

/******************************************************************************/
typedef struct CommandQueue {
    List *history; /* Command-list of commands from user interface */
    TSSemaphore sema;
} CommandQueue;

/******************************************************************************/
static void init_CommandQueue( CommandQueue *cmdlist ) {
    cmdlist->history = create_List( &type_Integer );
    cmdlist->sema = ts_create_semaphore();
}

/******************************************************************************/
static void destroy_CommandQueue( CommandQueue *cmdlist ) {
    destroy( cmdlist->history );
    ts_destroy_semaphore( cmdlist->sema );
}

/******************************************************************************/
static void enq_CommandQueue( CommandQueue *cmdlist, Command cmd )
{
    ts_close_semaphore( cmdlist->sema );
    append_List( r(cmdlist->history), r(create_Integer(cmd)) );
    ts_open_semaphore( cmdlist->sema );
#ifdef DEBUG_MCC
    printf("MCC: COMMAND %d is registered\n", cmd);
    fflush(stdout);
#endif
}

/******************************************************************************/
static bool isEmpty_CommandQueue( CommandQueue *cmdlist )
{
    bool res;

    ts_close_semaphore( cmdlist->sema );
    res = isEmpty_List( r(cmdlist->history) );
    ts_open_semaphore( cmdlist->sema );

    return res;
}

/******************************************************************************/
static Command deq_CommandQueue( CommandQueue *cmdlist )
{
    Integer *cmd;
    Command res = NOCMD;

    ts_close_semaphore( cmdlist->sema );
    if( !isEmpty_List( r(cmdlist->history) ) ) {
        cmd = (Integer *)get_List( r(cmdlist->history), 0 );
        res = (Command) value_Integer( r(cmd) );
        remove_List( r(cmdlist->history), 0 );
        destroy( cmd );
    }
    ts_open_semaphore( cmdlist->sema );

    return res;
}

/******************************************************************************/
static String *toString_CommandQueue( CommandQueue *cmdlist )
{
    StringBuffer *buf = NULL;
    int i, size;

    if( cmdlist == NULL ) return create_String( "NULL" );
    buf = create_StringBuffer();

    size = size_List( r(cmdlist->history) );
    appendFormat_StringBuffer( r(buf), "history: [%d]{", size );
    for( i = 0; i < size; i++ ) {
        Integer *icmd = get_List( r(cmdlist->history), i );
        if( i > 0 ) append_StringBuffer( r(buf), ", " );
        appendString_StringBuffer( r(buf), toString_Command( (Command)value_Integer( r(icmd) ) ) );
        destroy( icmd );
    }
    append_StringBuffer( r(buf), "}, ");
    appendFormat_StringBuffer( r(buf), "sema:%p", cmdlist->sema );

    return toString_Object(buf);
}

/******************************************************************************/
static void print_CommandQueue( CommandQueue *cmdlist )
{
    int i;

    printf("MCC: Commands:[\n");
    ts_close_semaphore( cmdlist->sema );
    for( i = 0; i < size_List( r(cmdlist->history) ); i++ ) {
        Integer *icmd = get_List( r(cmdlist->history), i );
        printf("       (%d):%d\n", i, value_Integer( r(icmd) ));
        destroy( icmd );
    }
    ts_open_semaphore( cmdlist->sema );
    printf("]\n");
    fflush(stdout);
}


/******************************************************************************/
/**                              TestRunStatus                               **/
/******************************************************************************/
typedef enum TestRunStatus {
      TESTRUN_FAILED = -1
    , TESTRUN_INITIAL
    , TESTRUN_CONFIGURED
    , TESTRUN_STARTED
    , TESTRUN_FINISHED
    , TESTRUN_REMOVED
} TestRunStatus;

/******************************************************************************/
static String *toString_TestRunStatus( TestRunStatus status )
{
    switch(status) {
        case TESTRUN_FAILED:     return create_String( "TESTRUN_FAILED" );
        case TESTRUN_INITIAL:    return create_String( "TESTRUN_INITIAL" );
        case TESTRUN_CONFIGURED: return create_String( "TESTRUN_CONFIGURED" );
        case TESTRUN_STARTED:    return create_String( "TESTRUN_STARTED" );
        case TESTRUN_FINISHED:   return create_String( "TESTRUN_FINISHED" );
        case TESTRUN_REMOVED:    return create_String( "TESTRUN_REMOVED" );
        defaulf:                 return format_String( "unknown(%d)", status );
    }
}


static String* process_nodelist_line ( ArrayList *nodes, const char *line );
static String* process_topology_line ( ArrayList *nodes, const char *line );
/******************************************************************************/
/**                                 TestRun                                  **/
/******************************************************************************/
typedef struct TestRun {
     String *testRunId;
     String *testName;
     String *parameters;
     String *numberOfNodes;
     String *topologyType;
     bool    save;
     String *prevTestRun;

     TestRunStatus status;
     ArrayList nodes;
     CommandQueue commands;

} TestRun;

/******************************************************************************/
static void init_TestRun( TestRun *testRun
                 , String *testRunId, String *testName,  String *parameters
                 , String* numberOfNodes, String *topologyType
                 , bool save, String *prevTestRun
                 )
{
    assertion( testRun != NULL, "init_TestRun: testRun == NULL" );

    testRun->testRunId = testRunId;
    testRun->testName = testName;
    testRun->parameters = parameters;
    testRun->numberOfNodes = numberOfNodes;
    testRun->topologyType = topologyType;
    testRun->save = save;
    testRun->prevTestRun = prevTestRun;

    testRun->status = TESTRUN_INITIAL;
    init_ArrayList( &testRun->nodes, sizeof(Node), NODE_ALLOC_BY, (Destructor)destroy_Node );
    init_CommandQueue( &testRun->commands );
}

/******************************************************************************/
static void destroy_TestRun( TestRun *testRun )
{
    assertion( testRun != NULL, "destroy_TestRun: testRun == NULL" );

    destroy( testRun->testRunId );
    destroy( testRun->testName );
    destroy( testRun->parameters );
    destroy( testRun->numberOfNodes );
    destroy( testRun->topologyType );
    destroy( testRun->prevTestRun );
    destroy_ArrayList( &testRun->nodes );
    destroy_CommandQueue( &testRun->commands );

}

typedef String *(*toString_func)(void *);
/*****************************  toString_ArrayList  ***************************/
static String *toString_ArrayList( ArrayList* array, toString_func toString_element )
{
    StringBuffer *buf = NULL;
    void* element;
    int i;

    assertion( (array != NULL), "toString_ArrayList: array is NULL" );
    assertion( (toString_element != NULL), "toString_ArrayList: toString_element is NULL" );

    buf = create_StringBuffer();
    appendFormat_StringBuffer( r(buf), "[%d]{", array->size );
    for( i = 0; i < array->size; i++ ) {
        element = get_ArrayList( array, i );
        if( i > 0 ) append_StringBuffer( r(buf), ", " );
        appendString_StringBuffer( r(buf), toString_element( element ) );
    }
    append_StringBuffer( r(buf), "}");

    return toString( buf );
}

/******************************************************************************/
static String *toString_TestRun( TestRun *testRun )
{
    StringBuffer *buf = NULL;

    if( testRun == NULL ) return create_String( "NULL" );
    buf = create_StringBuffer();
    append_StringBuffer( r(buf), "{" );

    append_StringBuffer( r(buf), "testRunId:" );
    appendString_StringBuffer( r(buf)
                             , testRun->testRunId != NULL ? r(testRun->testRunId)
                                                          : create_String( "NULL" )
                             );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "testName:" );
    appendString_StringBuffer( r(buf)
                             , testRun->testName != NULL ? r(testRun->testRunId)
                                                         : create_String( "NULL" )
                             );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "parameters:" );
    appendString_StringBuffer( r(buf)
                             , testRun->parameters != NULL ? r(testRun->parameters)
                                                           : create_String( "NULL" )
                             );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "computers:");
    appendString_StringBuffer( r(buf)
                             , testRun->numberOfNodes != NULL ? r(testRun->numberOfNodes)
                                                              : create_String( "NULL" )
                             );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "links:");
    appendString_StringBuffer( r(buf)
                             , testRun->topologyType != NULL ? r(testRun->topologyType)
                                                             : create_String( "NULL" )
                             );

    append_StringBuffer( r(buf), ", ");
    append_StringBuffer( r(buf), "prevTestRun:");
    appendString_StringBuffer( r(buf)
                             , testRun->prevTestRun != NULL ? r(testRun->prevTestRun)
                                                            : create_String( "NULL" )
                             );

    append_StringBuffer( r(buf), "}" );
    return toString_Object(buf);
}

/******************************************************************************/
static bool compareId_TestRun( TestRun *testRun, String *testRunId )
{
    assertion( testRun != NULL, "compareById_TestRun: testRun == NULL" );
    assertion( testRunId != NULL, "compareById_TestRun: testRunId == NULL" );

    return equals( r(testRun->testRunId), r(testRunId) );
}

/******************************************************************************/
static String* setNodes_TestRun( TestRun *testRun, const char* nodeListFile )
{
    String *res = NULL;
    FILE *file;
    char line[BUFF_SIZE];
    int lines = 0;

    assertion( testRun != NULL, "setNodes_TestRun: testRun == NULL" );

    file = fopen( nodeListFile, "r" );
    if( file == NULL ) {
        res = create_String( FILEOPEN_COORD_SETNODE_ERROR );
        return res;
    }

    while( fgets( line, LINE_SIZE, file ) != NULL ) {
        lines++;
        res = process_nodelist_line( &testRun->nodes, line );
        if( res != NULL ) {
            StringBuffer *buf = create_StringBuffer();
            appendFormat_StringBuffer( r(buf), "Line %d: ", lines);
            appendString_StringBuffer( r(buf), /* no r */res );
            return toString_Object(buf);
        }
    }

    if( EOF == fclose( file ) ) {
        res = create_String( FILECLOSE_COORD_SETNODE_ERROR );
        return res;
    }

    return NULL;
}

/******************************************************************************/
static String* setLinks_TestRun( TestRun *testRun, const char* topologyFile )
{
    String *res = NULL;
    FILE *file;
    char line[LINE_SIZE];
    int lines = 0;

    assertion( testRun != NULL, "setLinks_TestRun: testRun == NULL" );
    assertion( testRun->nodes.size != 0, "setLinks_TestRun: " NONODE_COORD_SETTOPOLOGY_ERROR);

    file = fopen( topologyFile, "r" );
    if( file == NULL ) {
        res = create_String( FILEOPEN_COORD_SETTOPOLOGY_ERROR );
        return res;
    }

    while( fgets( line, LINE_SIZE, file ) != NULL ) {
        lines++;
        res = process_topology_line( &testRun->nodes, line );
        if( res != NULL ) {
            StringBuffer *buf = create_StringBuffer();
            appendFormat_StringBuffer( r(buf), "Line %d: ", lines);
            appendString_StringBuffer( r(buf), /* no r */res );
            return toString_Object(buf);
        }
    }

    if( EOF == fclose( file ) ) {
        res = create_String( FILECLOSE_COORD_SETTOPOLOGY_ERROR );
        return res;
    }

    return NULL;
}

/******************************************************************************/
static int getNumberOfNodes_TestRun( TestRun *testRun )
{
    assertion( testRun != NULL, "getNumberOfNodes_TestRun: testRun == NULL" );

    return testRun->nodes.size;
}

/******************************************************************************/
static String *getNodeInfo_TestRun( TestRun *testRun, int nodeNumber )
{
    String *res = NULL;
    Node *node;

#ifdef DEBUG_MCC
    fprintf(stderr,"MCC: getNodeInfo_TestRun( %s, %d) count = %d\n", toCharArray_String( r(testRun->testRunId) ), nodeNumber, testRun->nodes.size );fflush(stdout);
#endif

    assertion( testRun != NULL, "getNodeInfo_TestRun: testRun == NULL" );
    assertion( nodeNumber < testRun->nodes.size, "getNodeInfo_TestRun: nodeNumber is out of range" );
    node = (Node*)get_ArrayList(&testRun->nodes, nodeNumber);

    res = clone_Object( r(node->info) );

    return res;
}

/******************************************************************************/
static String *getInfo_TestRun( TestRun *testRun )
{
    StringBuffer* buf  = create_StringBuffer();
    String *nodeInfo = NULL;
    int i;

    assertion( testRun != NULL, "getInfo_TestRun: testRun == NULL" );

    append_StringBuffer( r(buf), "{" );

    append_StringBuffer( r(buf), "\"name\":\"" );
    appendString_StringBuffer( r(buf)
                             , testRun->testName != NULL ? r(testRun->testName)
                                                         : create_String("NULL")
                             );

    append_StringBuffer( r(buf), "\"," );
    append_StringBuffer( r(buf), "\"params\":\"" );
    appendString_StringBuffer( r(buf)
                              , testRun->parameters != NULL ? r(testRun->parameters)
                                                            : create_String("NULL")
                             );

    append_StringBuffer( r(buf), "\"," );
    append_StringBuffer( r(buf), "\"nodeCount\":");
    appendString_StringBuffer( r(buf)
                             , testRun->numberOfNodes != NULL ? r(testRun->numberOfNodes)
                                                              : create_String("NULL")
                             );

    append_StringBuffer( r(buf), "," );
    append_StringBuffer( r(buf), "\"links\":\"");
    appendString_StringBuffer( r(buf)
                             , testRun->topologyType != NULL ? r(testRun->topologyType)
                                                             : create_String("NULL")
                             );

    append_StringBuffer( r(buf), "\"," );
    append_StringBuffer( r(buf), "\"log\":" );
    appendFormat_StringBuffer( r(buf), "%s", testRun->save ? "true" : "false" );

    append_StringBuffer( r(buf), ",");
    append_StringBuffer( r(buf), "\"prevTestRun\":\"");
    appendString_StringBuffer( r(buf)
                             , testRun->prevTestRun != NULL ? r(testRun->prevTestRun)
                                                            : create_String("")
                             );

    append_StringBuffer( r(buf), "\",");
    append_StringBuffer( r(buf), "\"status\":");
    appendString_StringBuffer( r(buf), to_JSON_String_NodeStatus( testRun->status ) );
    append_StringBuffer( r(buf), "}" );

    return toString_Object(buf);
}

/******************************************************************************/
static void printNodes_TestRun( TestRun *testRun )
{
    int i, s;

    assertion( testRun != NULL, "printNodes_TestRun: testRun == NULL" );

    for( i = 0; i < testRun->nodes.size; i++ ) {
        Node *node = (Node *)get_ArrayList( &testRun->nodes, i );
        printf( "MCC: %d: address:%s, port:%d, status:%d, out:["
              , i + 1
              , toCharArray_String( r(node->address) )
              , node->mccPort, node->status
              );
        if( node->outConections != NULL) {
            for( s = 0; s < size_Set( r(node->outConections) ); s++ ){
                String *S;
                int num;

                S = get_Set( r(node->outConections), s );
                num = atoi( toCharArray_String( r(S) ) );
                destroy( S );

                Node *tmpNode = (Node *)get_ArrayList( &testRun->nodes, num - 1 );
                printf( " %s:%d ", toCharArray_String( r(tmpNode->address) ), tmpNode->inPort );
            }
        }
        printf("], inPort:%d, in:[", node->inPort);
        if(node->inConections != NULL) {
            for( s = 0; s < size_Set( r(node->inConections) ); s++ ){
                String *S;
                int num;

                S = get_Set( r(node->inConections), s );
                num = atoi( toCharArray_String( r(S) ) );
                destroy( S );

                Node *tmpNode = (Node *)get_ArrayList( &testRun->nodes, num - 1 );
                printf( " %s:%d ", toCharArray_String( r(tmpNode->address) ), tmpNode->inPort );
            }
        }
        printf("]\n");
    }
}


/******************************************************************************/
/**                                  MccData                                 **/
/******************************************************************************/
typedef struct MccData {
    String        *config_dir;
    String        *cluster;
    String        *testruns_dir;
    bool          keep_connection;

    ArrayList     testRunList;
    TSSemaphore   info_sema;

#if defined(WIN32)
    HANDLE        thread;
#elif defined(unix)
    pthread_t     thread;
    time_t        thread_sleep_time;
    int           turns_left_for_info;
    volatile bool thread_death_required;
    volatile bool thread_death_completed;

#endif

} MccData;

/******************************************************************************/
static MccData MCC;

/******************************************************************************/
static void init_MccData( MccData *mcc )
{
    assertion( mcc != NULL, "init_MccData: mcc == NULL" );

    mcc->config_dir = NULL;
    mcc->cluster = NULL;
    mcc->testruns_dir = NULL;
    
    mcc->keep_connection = true;

    init_ArrayList( &mcc->testRunList, sizeof(TestRun), TESTRUN_ALLOC_BY, (Destructor)destroy_TestRun );
    mcc->info_sema = ts_create_semaphore();

    mcc->thread = 0;
    mcc->thread_sleep_time = DEFAULT_THREAD_SLEEPTIME;
    mcc->turns_left_for_info = DEFAULT_TURNS_LEFT_FOR_INFO;
    mcc->thread_death_required = false;
    mcc->thread_death_completed = false;

}

/******************************************************************************/
static void destroy_MccData( MccData *mcc )
{
    assertion( mcc != NULL, "destroy_MccData: mcc == NULL" );

    destroy( mcc->config_dir );
    destroy( mcc->cluster );
    destroy( mcc->testruns_dir );
    
    destroy_ArrayList( &mcc->testRunList );
    ts_destroy_semaphore( mcc->info_sema );
}

/******************************************************************************/
static String *registerCommand_MccData( MccData *mcc, String *testRunId, Command cmd )
{
    String *res = NULL;
    TestRun *testRun;

    assertion( mcc != NULL, "registerCommand_MccData: mcc == NULL" );
    assertion( testRunId != NULL, "registerCommand_MccData: testRunId == NULL" );
    testRun = (TestRun*) find_ArrayList( &mcc->testRunList, (ArrayListComparator)compareId_TestRun, testRunId );
    // TODO: return appropriate string if testRun is not found or test is not in correct state
    enq_CommandQueue( &testRun->commands, cmd );

    return res;
}

/******************************************************************************/
static String *getInfo_MccData( MccData *mcc )
{
    StringBuffer* buf  = create_StringBuffer();
    String *testRunInfo = NULL;
    int tr;

    append_StringBuffer( r(buf), "{" );
    for( tr = 0; tr < mcc->testRunList.size; tr++ ) {
        TestRun *testRun = (TestRun*)get_ArrayList( &mcc->testRunList, tr );

        if( tr > 0 ) append_StringBuffer( r(buf), "," );
        append_StringBuffer( r(buf), "\"" );
        appendString_StringBuffer( r(buf)
                                 ,  testRun->testRunId != NULL ? r(testRun->testRunId)
                                                               : create_String("NULL")
                                 );
        append_StringBuffer( r(buf), "\":" );

        testRunInfo = getInfo_TestRun( testRun );
        if( testRunInfo == NULL || length_String( r(testRunInfo) ) == 0 ) {
            append_StringBuffer( r(buf), "{}" );
        } else {
            appendString_StringBuffer( r(buf), r(testRunInfo) );
        }
        destroy( testRunInfo );
    }
    append_StringBuffer( r(buf), "}" );

    return toString_Object(buf);
}

/******************************************************************************/
/**                                 nodes.cfg                                **/
/******************************************************************************/
/* line format:
 *  address : mcc_port, sync_in_port
 */

typedef enum NodelineParserStates {
                                BEFORE_ADDRESS, ADDRESS, AFTER_ADDRESS
                              , BEFORE_FIRST_PORT, FIRST_PORT, AFTER_FIRST_PORT
                              , BEFORE_SECOND_PORT, SECOND_PORT, AFTER_SECOND_PORT
} NodelineParserStates;

/******************************************************************************/
/* returns:
 *   0 - wrong format
 *   1 - ok
 *  -1 - empty line
 */
static int parse_nodelist_line( const char *line, char *address, char *mccPort, char *inPort )
{
    int i, size;
    int startAddr = 0, endAddr = 0, startMccPort = 0, endMccPort = 0, startInPort = 0, endInPort = 0;
    NodelineParserStates state = BEFORE_ADDRESS;

    assertion( line != NULL, "parse_nodelist_line: line == NULL" );
    assertion( address != NULL, "parse_nodelist_line: address == NULL" );
    assertion( mccPort != NULL, "parse_nodelist_line: mccPort == NULL" );
    assertion( inPort != NULL, "parse_nodelist_line: inPort == NULL" );

    size = strlen( line );
    for( i=0; i < size; i++ ) {

        if( isspace(line[i]) ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                case AFTER_ADDRESS:
                case BEFORE_FIRST_PORT:
                case AFTER_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case AFTER_SECOND_PORT:
                    continue;
                case ADDRESS:
                    state = AFTER_ADDRESS;
                    endAddr = i - 1;
                    break;
                case FIRST_PORT:
                    state = AFTER_FIRST_PORT;
                    endMccPort = i - 1;
                    break;
                case SECOND_PORT:
                    state = AFTER_SECOND_PORT;
                    endInPort = i - 1;
                    break;
            }
        } else if( isalpha(line[i]) || line[i] == '_' || line[i] == '-' ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                    state = ADDRESS;
                    startAddr = i;
                    break;
                case ADDRESS:
                    continue;
                case AFTER_ADDRESS:
                case BEFORE_FIRST_PORT:
                case FIRST_PORT:
                case AFTER_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case SECOND_PORT:
                case AFTER_SECOND_PORT:
                    return 0; /* wrong format */
            }
        } else if( isdigit(line[i]) ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                    state = ADDRESS;
                    startAddr = i;
                    break;
                case ADDRESS:
                case FIRST_PORT:
                case SECOND_PORT:
                    continue;
                case AFTER_ADDRESS:
                case AFTER_FIRST_PORT:
                case AFTER_SECOND_PORT:
                    return 0; /* wrong format */
                case BEFORE_FIRST_PORT:
                    state = FIRST_PORT;
                    startMccPort = i;
                    break;
                case BEFORE_SECOND_PORT:
                    state = SECOND_PORT;
                    startInPort = i;
                    break;
            }
        }
        else if ( line[i] == '.' ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                case AFTER_ADDRESS:
                case BEFORE_FIRST_PORT:
                case FIRST_PORT:
                case AFTER_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case SECOND_PORT:
                case AFTER_SECOND_PORT:
                        return 0; /* wrong format */
                case ADDRESS:
                    continue;
            }
        } else if ( line[i] == ':' ) {
            switch( state ) {
                case BEFORE_ADDRESS:
                case BEFORE_FIRST_PORT:
                case FIRST_PORT:
                case AFTER_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case SECOND_PORT:
                case AFTER_SECOND_PORT:
                    return 0;  /* wrong format */
                case ADDRESS:
                    state = BEFORE_FIRST_PORT;
                    endAddr = i - 1;
                    break;
                case AFTER_ADDRESS:
                    state = BEFORE_FIRST_PORT;
                    break;
            }
        } else if ( line[i] == ',') {
            switch( state ) {
                case BEFORE_ADDRESS:
                case ADDRESS:
                case AFTER_ADDRESS:
                case BEFORE_FIRST_PORT:
                case BEFORE_SECOND_PORT:
                case SECOND_PORT:
                case AFTER_SECOND_PORT:
                    return 0;
                case FIRST_PORT:
                    state = BEFORE_SECOND_PORT;
                    endMccPort = i - 1;
                    break;
                case AFTER_FIRST_PORT:
                    state = BEFORE_SECOND_PORT;
                    break;
            }
        } else {
            return 0;
        }

    } /* for */
    if( state == SECOND_PORT ) {
        endInPort = size;
        state = AFTER_SECOND_PORT;
    }
    if( state == BEFORE_ADDRESS ) return -1; /* empty line */
    if( state != AFTER_SECOND_PORT ) return 0;

    memcpy( address, line + startAddr, endAddr - startAddr + 1 );
    address[endAddr - startAddr + 1] = 0;
    memcpy( mccPort, line + startMccPort, endMccPort - startMccPort + 1 );
    mccPort[endMccPort - startMccPort + 1] = 0;
    memcpy( inPort, line + startInPort, endInPort - startInPort + 1 );
    inPort[endInPort - startInPort + 1] = 0;

    return 1;
}

/******************************************************************************/
static String* process_nodelist_line ( ArrayList *nodes, const char *line )
{
    String *res = NULL;
    int rc;
    char address[ADDRESS_SIZE], str_mccPort[PORT_SIZE], str_inPort[PORT_SIZE];
    unsigned short  mccPort, inPort;

    assertion( line != NULL, "parse_nodelist_line: line == NULL" );
#ifdef DEBUG_MCC
    printf("MCC: process_nodelist_line, line:<%s>\n", line);fflush(stdout);
#endif

    rc = parse_nodelist_line( line, address, str_mccPort, str_inPort );
     if( rc == 0 ) {
        res = create_String( FORMAT_COORD_SETNODE_ERROR );
        return res;
     }
     if( rc != -1 ) { /* line is not empty */
         Node *node = (Node *)append_ArrayList( nodes );
         init_Node( node
                  , create_String(address)
                  , atoi( str_mccPort )
                  , atoi( str_inPort )
                  );
     }

     return NULL;

}

/******************************************************************************/
/**                                 links.cfg                                **/
/******************************************************************************/
/* line format:
 *  node : node1, node2, node3....
 */

typedef enum TopologylineParserStates {
                                BEFORE_NODE_NUMBER, NODE_NUMBER, AFTER_NODE_NUMBER
                              , BEFORE_NODE_IN_LIST, NODE_IN_LIST, AFTER_NODE_IN_LIST
} TopologylineParserStates;

/******************************************************************************/
/* returns
 *   0 - wrong format
 *   1 - ok
 *  -1 - empty line
 */
static int parse_topology_line( const char *line, char *node_number, List **out_nodes_list )
{
    int i, size;
    int startToken = 0;
    char out_node[NUMBER_SIZE];
    TopologylineParserStates state = BEFORE_NODE_NUMBER;

    assertion( line != NULL, "parse_topology_line: line is NULL" );
    assertion( node_number != NULL, "parse_topology_line: node_number is NULL" );
    assertion( out_nodes_list != NULL, "parse_topology_line: out_nodes_list is NULL" );

#ifdef DEBUG_MCC
    printf("MCC: parse_topology_line, line:<%s>\n", line);
#endif
    size = strlen( line );
    for( i=0; i < size; i++ ) {

        if( isspace(line[i]) ) {
            switch( state ) {
                case BEFORE_NODE_NUMBER:
                case AFTER_NODE_NUMBER:
                case BEFORE_NODE_IN_LIST:
                case AFTER_NODE_IN_LIST:
                    continue;
                case NODE_NUMBER:
                    state = AFTER_NODE_NUMBER;
                    memcpy( node_number, line + startToken, i - startToken );
                    node_number[i - startToken] = 0;
                    break;
                case NODE_IN_LIST:
                    state = AFTER_NODE_IN_LIST;
                    memcpy( out_node, line + startToken, i - startToken );
                    out_node[i - startToken] = 0;
                    append_List( r(*out_nodes_list), r(create_String(out_node)) );
                    break;
            }
        } else if( isdigit(line[i]) ) {
            switch( state ) {
                case BEFORE_NODE_NUMBER:
                    state = NODE_NUMBER;
                    startToken = i;
                    break;
                case NODE_NUMBER:
                case NODE_IN_LIST:
                    continue;
                case AFTER_NODE_NUMBER:
                case AFTER_NODE_IN_LIST:
                    return 0;
                case BEFORE_NODE_IN_LIST:
                    state = NODE_IN_LIST;
                    startToken = i;
                    break;
            }
        } else if ( line[i] == ':') {
            switch( state ) {
                case BEFORE_NODE_NUMBER:
                case BEFORE_NODE_IN_LIST:
                case NODE_IN_LIST:
                case AFTER_NODE_IN_LIST:
                    return 0;
                case NODE_NUMBER:
                    state = BEFORE_NODE_IN_LIST;
                    memcpy( node_number, line + startToken, i - startToken );
                    node_number[i - startToken] = 0;
                    break;
                case AFTER_NODE_NUMBER:
                    state = BEFORE_NODE_IN_LIST;
                    break;
            }
        } else if ( line[i] == ',') {
            switch( state ) {
                case BEFORE_NODE_NUMBER:
                case NODE_NUMBER:
                case AFTER_NODE_NUMBER:
                case BEFORE_NODE_IN_LIST:
                    return 0;
                case NODE_IN_LIST:
                    state = BEFORE_NODE_IN_LIST;
                    memcpy( out_node, line + startToken, i - startToken );
                    out_node[i - startToken] = 0;
                    append_List( r(*out_nodes_list), r(create_String(out_node)) );
                    break;
                case AFTER_NODE_IN_LIST:
                    state = BEFORE_NODE_IN_LIST;
                    break;
            }
        } else {
            return 0;
        }

    } /* for */
    if( state == NODE_IN_LIST) {
        memcpy( out_node, line + startToken, size - startToken + 1);
        out_node[size - startToken + 1] = 0;
        append_List( r(*out_nodes_list), r(create_String(out_node)) );
        state = AFTER_NODE_IN_LIST;
    }
    if( state == BEFORE_NODE_NUMBER ) return -1; /* empty line */
    if( state != AFTER_NODE_IN_LIST ) return 0;

    return 1;
}

/******************************************************************************/
static String* process_topology_line ( ArrayList *nodes, const char *line )
{
    String *res = NULL;
    int rc;
    char str_node_number[NUMBER_SIZE];
    List *link_list = create_List( &type_String );

    assertion( line != NULL, "process_topology_line: line == NULL" );

    rc = parse_topology_line( line, str_node_number, &link_list );
    if( rc == 0 ) {
        res = format_String( FORMAT_COORD_SETTOPOLOGY_ERROR );
        goto ERROR;
    }
    if( rc != -1 ) { /* line is not empty */
        int node;

        node = atoi( str_node_number );
        if( node > nodes->size ) {
            res = format_String( RANGE_COORD_SETTOPOLOGY_ERROR, node, nodes->size );
            goto ERROR;
        }

        while (! isEmpty_List( r(link_list) ) ) {
            String *cur_node = get_List( r(link_list), 0 );
            const char *tmp_str_number = toCharArray_String( r(cur_node) );
            int tmp_number = atoi( tmp_str_number );
            Node *tmpNode = NULL;

            remove_List( r(link_list), 0 );
            if( contains_List( r(link_list), r(cur_node) ) ) {
                res = format_String( DUPLICATE_COORD_SETTOPOLOGY_ERROR, tmp_str_number );
                goto ERROR;
            }
            tmpNode = (Node*)get_ArrayList( nodes, node - 1 );
            add_Set( r(tmpNode->outConections),  /* no r */cur_node );

            tmpNode = (Node*)get_ArrayList( nodes, tmp_number - 1 );
            add_Set( r(tmpNode->inConections), r(create_String(str_node_number)) );
        }
    }

ERROR:
    destroy(link_list);
    return res;
}

/******************************************************************************/
/**                             coord_saveLog                                **/
/******************************************************************************/
String* coord_saveLog( int node, const char* fileName )
{
    String *res = NULL;

/*

    if( numOfNodes == 0 ) {
        res = create_String( NONODE_COORD_LOG_ERROR );
        return res;
    }

    if( fileName == NULL ) {
        res = create_String( NULL_COORD_LOG_ERROR );
        return res;
    }

    if(node > numOfNodes) {
        res = format_String( RANGE_COORD_LOG_ERROR,
                             node, numOfNodes );
        return res;
    }

    nodes[node].saveLog = true;
    nodes[node].logFileName = create_String( fileName );
*/
/*
nodeNumber = node;
saveLog = true;
if(logFileName != NULL) {
    destroy(logFileName);
}
logFileName = create_String( fileName );
*/
    return res;
}

/******************************************************************************/
String* coord_unSaveLog( int node )
{
    String *res = NULL;
/*
    if( numOfNodes == 0 ) {
        res = create_String( NONODE_COORD_LOG_ERROR );
        return res;
    }

    if(node > numOfNodes) {
        res = format_String( RANGE_COORD_LOG_ERROR,
                             node, numOfNodes );
        return res;
    }

    nodes[node].saveLog = false;
*/
    /*
nodeNumber = -1;
saveLog = false;
if(logFileName != NULL) {
    destroy(logFileName);
    logFileName = NULL;
}
*/
    return res;
}

/******************************************************************************/
/**                             coord_loadLog                                **/
/******************************************************************************/
String* coord_loadLog( int node, const char* fileName, bool tA )
{
    String *res = NULL;

/*    if( numOfNodes == 0 ) {
        res = create_String( NONODE_COORD_LOG_ERROR );
        return res;
    }

    if( fileName == NULL ) {
        res = create_String( NULL_COORD_LOG_ERROR );
        return res;
    }

    if(node > numOfNodes) {
        res = format_String( RANGE_COORD_LOG_ERROR,
                             node, numOfNodes );
        return res;
    }

    nodes[node].loadLog = true;
    nodes[node].logFileName = create_String( fileName );
    nodes[node].thenAppend = thenAppend;
*/
/*
nodeNumber = node;
saveLog = false;
loadLog = true;
if(logFileName != NULL) {
    destroy(logFileName);
}
logFileName = create_String( fileName );
thenAppend = tA;
*/
    return res;
}


/******************************************************************************/
/**                   send command & config & get info                       **/
/******************************************************************************/
static int coord_send_command(const char *address, unsigned short  port, const char *command, char *answer, int msg_size)
{
int rc;
int sock = -1;

#ifdef DEBUG_MCC
    printf("MCC: connecting to %s:%d...\n", address, port);
#endif
    sock = connect_to_server( address, port );
    if( sock == -1 ) return -1;
#ifdef DEBUG_MCC
    printf("MCC: -->sending <%s> ...\n", command);
    fflush(stdout);
#endif
    rc = send_to_socket(sock, command, msg_size);
#ifdef DEBUG_MCC
    printf("MCC: <--sent:%d <%s> ...\n", rc, command);
    fflush(stdout);
#endif
    rc = read_from_socket(sock, answer, msg_size);
#ifdef DEBUG_MCC
    printf("MCC: <--answer:%d <%s> ...\n", rc, answer);
    fflush(stdout);
#endif
    close_socket(sock);

    return rc;
}

/******************************************************************************/
static int coord_send_config(ArrayList *nodes, int cpu_id, char *answer, int msg_size)
{
    int rc, i;
    int sock = -1;
    char message[BUFF_SIZE];
    Node *node = (Node*)get_ArrayList( nodes, cpu_id );
    StringBuffer *buf = NULL;
    String *id = NULL;

#ifdef DEBUG_MCC
    printf( "MCC: send_config: connecting to %s:%d...\n", toCharArray_String( r(node->address) ), node->mccPort );
    fflush(stdout);
#endif
    sock = connect_to_server( toCharArray_String( r(node->address) ), node->mccPort );
    if( sock != -1 ) {
        rc = send_to_socket( sock, MCC_CMD_CFG, msg_size );

        buf = create_StringBuffer();
        //appendFormat_StringBuffer( r(buf), "(%03d)", cpu_id );
        appendString_StringBuffer( r(buf), r(node->address) );
        append_StringBuffer( r(buf), ":" );
        appendString_StringBuffer( r(buf), valueOfInt_String( node->mccPort ) );
        id = toString_Object( /* no r */buf );

        sprintf( message, MCC_CMD_CFG_INFO
               , cpu_id, node->inPort
               , size_Set( r(node->inConections) )
               , toCharArray_String( r(id) )
               );
        destroy( id );

        rc = send_to_socket( sock, message, msg_size );

        for( i = 0; i < size_Set( r(node->outConections) ); i++ ) {
            String *S;
            int num;
            Node *tmpNode;

            S = get_Set( r(node->outConections), i );
            num = atoi( toCharArray_String(r(S)) );

            tmpNode = (Node*)get_ArrayList( nodes, num - 1 );
            sprintf( message, MCC_CMD_CFG_OUT
                   , tmpNode->inPort, toCharArray_String( r(tmpNode->address) )
                   );
#ifdef DEBUG_MCC
            printf( "MCC: send_config: sending <%s>...\n", message );
            fflush(stdout);
#endif
            rc = send_to_socket( sock, message, msg_size );
        }

        rc = send_to_socket( sock, MCC_CMD_CFG_END, msg_size );


        sprintf( message, MCC_CMD_CFG_LOG,
                 node->saveLog,
                 node->loadLog,
                 node->logFileName == NULL ? "NULL" : toCharArray_String( r(node->logFileName) )
                 );

/*        if(node_number == nodeNumber) {
         sprintf( message, MCC_CMD_CFG_LOG,
             saveLog,
             loadLog,
             thenAppend,
             logFileName == NULL ? "NULL" : toCharArray_String( r(logFileName) )
             );
        }
        */
#ifdef DEBUG_MCC
    printf("MCC: send_config: sending <%s>...\n", message);
#endif
        rc = send_to_socket( sock, message, msg_size );

        rc = read_from_socket( sock, answer, msg_size );

        close_socket( sock );

        return rc;
    } else return -1;

}


/******************************************************************************/
static String *recvInfo_Node( Node *node )
{
    String *res = NULL;
    int rc = -1;
    int sock = -1;
    int len;
    char answer[AGENT_MSG_SIZE];
    char *buffer;

    assertion( node != NULL, "recvInfo_Node: node == NULL" );

#ifdef DEBUG_MCC
    printf("MCC: recvInfo_Node: connecting to %s:%d...\n", toCharArray_String( r(node->address) ), node->mccPort ); fflush(stdout);
#endif
    sock = connect_to_server( toCharArray_String( r(node->address) ), node->mccPort );
    if( sock != -1 ) {
#ifdef DEBUG_MCC
        printf("MCC: recvInfo_Node: -> sending <%s> ...\n", MCC_CMD_INFO); fflush(stdout);
#endif
        rc = send_to_socket(sock, MCC_CMD_INFO, AGENT_MSG_SIZE);

        /* get length of string*/
        rc = read_from_socket(sock, answer, AGENT_MSG_SIZE);
#ifdef DEBUG_MCC
        printf("MCC: recvInfo_Node: <- answer:<%s> ...\n", answer); fflush(stdout);
#endif
        sscanf( answer, "%d", &len) ;
        buffer = (char *) calloc( len + 1, sizeof(char) );

        /* get string itself */
        rc = read_from_socket( sock, buffer, len + 1 );
#ifdef DEBUG_MCC
        printf("MCC: recvInfo_Node: <- buffer:<%s> ...\n", buffer); fflush(stdout);
#endif

        close_socket( sock );

        res = create_String( buffer );
        free( buffer );
    }

     return res;
}


static String *startTest_Agent_TestRun( TestRun *testRun );
static String *stopTest_Agent_TestRun( TestRun *testRun );
static String *getEndallScriptName();
static String *createParamStringForKill(String *testRunId);
static String *runScript( String *scriptName, String *scriptParams );
/******************************************************************************/
/**                             GetInfo thread                               **/
/******************************************************************************/
static void *thread_GetInfo( void *par )
{
    int i, tr;
    int turns = MCC.turns_left_for_info;
    String *endall = NULL, *params = NULL;

#ifdef DEBUG_MCC
    printf("MCC: GetInfo thread started\n"); fflush(stdout);
#endif

    while( !MCC.thread_death_required ) {
        for( tr = 0; tr < MCC.testRunList.size; tr++ ) {
            TestRun *testRun = (TestRun*)get_ArrayList( &MCC.testRunList, tr );

            /* Check command queue for a command */
            if( ! isEmpty_CommandQueue( &testRun->commands ) ) {
                Command cmd = deq_CommandQueue( &testRun->commands );
#ifdef DEBUG_MCC
                printf("MCC: thread_GetInfo: executing command %d...\n", cmd); fflush(stdout);
#endif
                switch(cmd) {
                case START:
                    startTest_Agent_TestRun( testRun );
                    break;
    //            case STOPTEST:
    //                agent_StopTest();
    //                break;
                case STOP:
                    stopTest_Agent_TestRun( testRun );
                    // TODO
                    // run endall.sh
                    endall = getEndallScriptName();
                    params = createParamStringForKill(testRun->testRunId);
                    runScript( endall, params );
                    //destroy( endall );
                    //destroy( params );
                    break;
                }
            }
        }
        if( !turns-- ) {
#ifdef DEBUG_MCC
            printf("MCC: thread_GetInfo: getting info...\n"); fflush(stdout);
#endif
            turns = MCC.turns_left_for_info;
            for( tr = 0; tr < MCC.testRunList.size; tr++ ) {
                TestRun *testRun = (TestRun*)get_ArrayList( &MCC.testRunList, tr );
                for( i = 0; i < testRun->nodes.size; i++ ) {
                    String *res = NULL;
                    Node *node = (Node*)get_ArrayList( &testRun->nodes, i );
                    if( node->status != TEST_STOPPED ) {
                        res = recvInfo_Node( node );

                        ts_close_semaphore( MCC.info_sema );
                        {
                            if( node->info != NULL ) {
                                destroy( node->info );
                            }
                            node->info = res;
                        }
                        ts_open_semaphore( MCC.info_sema );
                    }
                }
            }
        }
        sleep_msec( MCC.thread_sleep_time );
    }

    MCC.thread_death_completed = true;

#ifdef DEBUG_MCC
    printf("MCC: GetInfo thread finished\n"); fflush(stdout);
#endif
    return NULL;
}


/***************************  start_GetInfoThread  ****************************/
#if defined(WIN32)

static DWORD WINAPI ThreadGetInfo(LPVOID lpParam)
{
    thread_GetInfo( (void*)lpParam );
    ExitThread(0);
    return 0;
}
static int start_GetInfoThread()
{
    DWORD dwThreadId;

    if (thread != 0) {
        fprintf( stderr, "MCC: GetInfo thread already started!\n" );
        return -1;
    }

    thread = CreateThread(NULL, 0, ThreadGetInfo, thread_GetInfo, 0, &dwThreadId);
    if (thread == 0) {
        printf("MCC: Creating of GetInfo thread failed: %d\n", GetLastError());
        return -1;
    }

    sleep_msec(1);
    return 0;
}

#elif defined(unix)

/******************************************************************************/
static int start_GetInfoThread()
{
    if( MCC.thread != 0 ) {
        fprintf( stderr, "MCC: GetInfo thread already started!\n" );
        return -1;
    }

    if( pthread_create( &MCC.thread, NULL, thread_GetInfo, NULL) != 0 ) {
        perror( "MCC: pthread_create" );
        return -1;
    }

    sleep_msec(1);
    return 0;
}
#endif


/******************************************************************************/
static int stop_GetInfoThread()
{
    int res;

    if( MCC.thread == 0 )
        return 0;

    MCC.thread_death_required = true;

    while (!MCC.thread_death_completed)
        sleep_msec( 10 );


#if defined(WIN32)
    for (;;) {
        DWORD exitCode;
        res = ! GetExitCodeThread( MCC.thread, &exitCode );
        if (exitCode == STILL_ACTIVE)
            sleep_msec( 10 );
        else break;
    }
    res = ! CloseHandle( MCC.thread ) || res;    // 0 iff CloseHandle()==true && GetExitCodeThread==true
#elif defined(unix)
    res = pthread_join( MCC.thread, NULL );
    if (res != 0)
        perror( "pthread_join" );
#endif

    MCC.thread = 0;
    return res;
}


/******************************************************************************/
/**                                                                          **/
/******************************************************************************/
String* configureNodes_TestRun( TestRun *testRun )
{
    String *res = NULL;
    char answer[AGENT_MSG_SIZE];
    int i;
    int numOfConnected = 0;
    int rc;
    int status;
    int attempts_left = DEFAULT_NUMBER_OF_ATTEMPTS;


    /* Will use the first node in nodes.cfg for testbench.fsm */
    if(testRun->nodes.size) {
        Node *node = (Node*)get_ArrayList( &testRun->nodes, 0 );
        if(testRun->save) {
            node->saveLog = testRun->save;
            node->logFileName = create_String(LOG_FILENAME);
        }
        if(testRun->prevTestRun != NULL) {
            node->loadLog = true;
            node->logFileName = create_String(LOG_FILENAME);
        }
    }

#ifdef DEBUG_MCC
    {
        String *info = toString_TestRun( testRun );
        printf( "MCC: testRun:%s\n", toCharArray_String( r(info) ) );
        destroy(info);
    }
#endif

    while( numOfConnected < testRun->nodes.size && attempts_left--) {
#ifdef DEBUG_MCC
        printf("MCC: startTest_Agent_TestRun: Trying to connect (%d attempt(s) left)...\n", attempts_left); fflush(stdout);
#endif
        for( i = 0; i < testRun->nodes.size; i++ ){
            Node *node = (Node*)get_ArrayList( &testRun->nodes, i );
            if( node->status == NODE_INITIAL ) {
                rc = coord_send_config( &testRun->nodes, i, answer, AGENT_MSG_SIZE );

                if( rc != -1 && strcmp(answer, AGENT_ANSWER_OK MCC_CMD_CFG) == 0) {
                    node->status = NODE_CONFIGURED;
                }

            }
        }
        numOfConnected = 0;
        for( i = 0; i < testRun->nodes.size; i++ ){
            Node *node = (Node*)get_ArrayList( &testRun->nodes, i );
            numOfConnected += (node->status == NODE_CONFIGURED);
        }

        if( numOfConnected < testRun->nodes.size ) sleep_msec(DEFAULT_TIMEOUT_BETWEEN_ATTEMPTS);
    }

    if( numOfConnected == testRun->nodes.size ) {
        testRun->status = TESTRUN_CONFIGURED;
    } else {
        res = create_String(CONNECT_MCC_ERROR);
        return res;
    }

    return res;
}


/******************************************************************************/
/**                             startTest_MCC                                **/
/******************************************************************************/
static String *startTest_TestRun( TestRun *testRun )
{
    String *res = NULL;

    assertion( testRun != NULL, "startTest_TestRun: testRun == NULL" );
    if( testRun->status == TESTRUN_CONFIGURED) {
        enq_CommandQueue( &testRun->commands, START );
    } else {
        res = create_String(START_MCC_ERROR);
    }

    return res;
}

/******************************************************************************/
static String *startTest_Agent_TestRun( TestRun *testRun )
{
    String *res = NULL;
    int i, rc = -1;
    char answer[AGENT_MSG_SIZE];
    int numOfConnected = 0;
    int attempts_left = DEFAULT_NUMBER_OF_ATTEMPTS;


    assertion( testRun != NULL, "agent_StartAll: testRun == NULL" );

    while( numOfConnected < testRun->nodes.size && attempts_left--) {
#ifdef DEBUG_MCC
        printf("MCC: startTest_Agent_TestRun: Trying to connect (%d attempt(s) left)...\n", attempts_left); fflush(stdout);
#endif
        for( i = 0; i < testRun->nodes.size; i++) {
            Node *node = (Node*)get_ArrayList(&testRun->nodes, i);

            rc =  coord_send_command( toCharArray_String( r(node->address) )
                                    , node->mccPort
                                    , MCC_CMD_START, answer, AGENT_MSG_SIZE
                                    );
            if( rc != -1 ) {
                if( strcmp( answer, AGENT_ANSWER_OK MCC_CMD_START ) == 0 ) {
                    node->status = TEST_STARTED;
                }
            }
        }
        numOfConnected = 0;
        for( i = 0; i < testRun->nodes.size; i++ ){
            Node *node = (Node*)get_ArrayList( &testRun->nodes, i );
            numOfConnected += (node->status == TEST_STARTED);
        }

        if( numOfConnected < testRun->nodes.size ) sleep_msec(DEFAULT_TIMEOUT_BETWEEN_ATTEMPTS);
    }

    if( numOfConnected == testRun->nodes.size ) {
        testRun->status = TESTRUN_STARTED;
    } else {
        res = create_String(CONNECT_MCC_ERROR);
        return res;
    }

    return res;
}

/******************************************************************************/
String *startTest_MCC( String *testRunId )
{
    String *res = NULL;
    TestRun *testRun;

    assertion( testRunId != NULL, "startTest_MCC: testRunId == NULL" );
    testRun = (TestRun*) find_ArrayList( &MCC.testRunList, (ArrayListComparator)compareId_TestRun, testRunId );
    if( testRun != NULL ) {
        res = startTest_TestRun( testRun );
    } else {
        res = create_String( NO_TESTRUN_MCC_ERROR );
    }

    destroy( testRun );
    return res;
}

/******************************************************************************/
/**                               stopTest_MCC                               **/
/******************************************************************************/
static String *stopTest_TestRun( TestRun *testRun )
{
    String *res = NULL;

    assertion( testRun != NULL, "stopTest_TestRun: testRun == NULL" );
    //if( testRun->status == TESTRUN_STARTED) {
        enq_CommandQueue( &testRun->commands, STOP );
    //} else {
    //    res = create_String(STOP_MCC_ERROR);
    //}

    return res;
}

/******************************************************************************/
static String *stopTest_Agent_TestRun( TestRun *testRun )
{
    String *res = NULL;
    int i, rc = -1;
    char answer[AGENT_MSG_SIZE];
    int numOfConnected = 0;
    int attempts_left = DEFAULT_NUMBER_OF_ATTEMPTS;


    assertion( testRun != NULL, "stopTest_Agent_TestRun: testRun == NULL" );

    while( numOfConnected < testRun->nodes.size && attempts_left--) {
#ifdef DEBUG_MCC
        printf("MCC: stopTest_Agent_TestRun: Trying to connect (%d attempt(s) left)...\n", attempts_left); fflush(stdout);
#endif
        for( i = 0; i < testRun->nodes.size; i++) {
            Node *node = (Node*)get_ArrayList(&testRun->nodes, i);

            rc =  coord_send_command( toCharArray_String( r(node->address) )
                                    , node->mccPort
                                    , MCC_CMD_STOP, answer, AGENT_MSG_SIZE
                                    );
            if( rc != -1 ) {
                if( strcmp( answer, AGENT_ANSWER_OK MCC_CMD_STOP ) == 0 ) {
                    node->status = TEST_STOPPED;
                }
            }
        }

        numOfConnected = 0;
        for( i = 0; i < testRun->nodes.size; i++ ){
            Node *node = (Node*)get_ArrayList( &testRun->nodes, i );
            numOfConnected += (node->status == TEST_STOPPED);
        }

        if( numOfConnected < testRun->nodes.size ) sleep_msec(DEFAULT_TIMEOUT_BETWEEN_ATTEMPTS);
    }


    if( numOfConnected == testRun->nodes.size ) {
        testRun->status = TESTRUN_FINISHED;
    } else {
        res = create_String(CONNECT_MCC_ERROR);
        return res;
    }

    return res;
}

/******************************************************************************/
String *coord_StopTest( String *testRunId ) //TODO: change name to stopTest_MCC
{
    String *res = NULL;
    TestRun *testRun;

    assertion( testRunId != NULL, "stopTest_MCC: testRunId == NULL" );
    testRun = (TestRun*) find_ArrayList( &MCC.testRunList, (ArrayListComparator)compareId_TestRun, testRunId );
    if( testRun != NULL ) {
        res = stopTest_TestRun( testRun );
    } else {
        res = create_String( NO_TESTRUN_MCC_ERROR );
    }

    destroy( testRun );
    return res;
}

/******************************************************************************/
/**                         coord_GetActiveTestRunIds                        **/
/******************************************************************************/
List   *coord_GetActiveTestRunIds() //TODO: change to getActiveTestRunIds_MCC
{
    int tr;
    List *ids = create_List( &type_String );

    for( tr = 0; tr < MCC.testRunList.size; tr++ ) {
        TestRun *testRun = (TestRun*)get_ArrayList( &MCC.testRunList, tr );
        append_List( r(ids), r(testRun->testRunId) );
    }

    return ids;
}

/******************************************************************************/
/**                          coord_GetNumberOfNodes                          **/
/******************************************************************************/
int coord_GetNumberOfNodes( String *testRunId ) //TODO: change to getNumberOfNodes_MCC
{
    TestRun *testRun;
    int numberOfNodes = 0;

    assertion( testRunId != NULL, "coord_GetNumberOfNodes: testRunId == NULL" );

    testRun = (TestRun*) find_ArrayList( &MCC.testRunList, (ArrayListComparator)compareId_TestRun, testRunId );
    if( testRun == NULL ) return -1;

    ts_close_semaphore( MCC.info_sema );
    numberOfNodes = getNumberOfNodes_TestRun( testRun );
    ts_open_semaphore( MCC.info_sema );

    return numberOfNodes;

}

/******************************************************************************/
/**                            coord_GetNodeInfo                             **/
/******************************************************************************/
String *coord_GetNodeInfo( String *testRunId, int nodeNumber ) //TODO: change to getNodeInfo_MCC
{
    String *res = NULL;
    TestRun *testRun;

    assertion( testRunId != NULL, "coord_GetNodeInfo: testRunId == NULL" );
    testRun = (TestRun*) find_ArrayList( &MCC.testRunList, (ArrayListComparator)compareId_TestRun, testRunId );
    if( testRun == NULL ) return NULL;

    ts_close_semaphore( MCC.info_sema );
    res = getNodeInfo_TestRun( testRun, nodeNumber );
    ts_open_semaphore( MCC.info_sema );

    return res;
}

/******************************************************************************/
/**                          coord_GetTestRunInfo                            **/
/******************************************************************************/
String *coord_GetTestRunInfo( String *testRunId ) //TODO: change to getTestRunInfo_MCC
{
    String *res = NULL;
    TestRun *testRun;

    assertion( testRunId != NULL, "coord_GetTestRunInfo: testRunId == NULL" );
    testRun = (TestRun*) find_ArrayList( &MCC.testRunList, (ArrayListComparator)compareId_TestRun, testRunId );
    if( testRun == NULL ) return NULL;

    ts_close_semaphore( MCC.info_sema );
    res = getInfo_TestRun( testRun );
    ts_open_semaphore( MCC.info_sema );

    return res;
}

/******************************************************************************/
/**                            coord_GetInfoAll                              **/
/******************************************************************************/
String *coord_GetInfoAll() //TODO: change to getInfo_MCC
{
    String *res = NULL;

    ts_close_semaphore( MCC.info_sema );
    res = getInfo_MccData( &MCC );
    ts_open_semaphore( MCC.info_sema );

    return res;
}

/******************************************************************************/
/**                                                                          **/
/******************************************************************************/


const char* coord_getConfigDir()
{
    return toCharArray_String( r(MCC.config_dir) );
}

const char* coord_getCluster()
{
    return toCharArray_String( r(MCC.cluster) );
}

const char* coord_getTestRunsDir()
{
    return toCharArray_String( r(MCC.testruns_dir) );
}

/******************************************************************************/
static String* set_cluster( const char* name )
{
    String* res = NULL;
    String* clusters_dir_string = format_String( "%s/%s"
                                               , coord_getConfigDir()
                                               , CLUSTERS_DIR
                                               );
    const char* clusters_dir = toCharArray_String( r(clusters_dir_string) );
    DIR *dir = opendir( clusters_dir );
    struct dirent *ent;

    if( dir == NULL )
    {
        printf( "MCC: Cluster configurations dir '%s' not found.\n", clusters_dir );
        destroy( clusters_dir_string );
        return false;
    }
    while( (ent = readdir( dir ) ) != NULL )
    {
        if( /*ent->d_type == DT_DIR &&*/ ent->d_name[0] != '.' )
        {
            if( name != NULL && strcmp( ent->d_name, name ) != 0 ) continue;
            MCC.cluster = create_String( ent->d_name );
            break;
        }
    }
    closedir( dir );
    if( MCC.cluster == NULL )
    {
        if( name == NULL )
        {
            res = format_String( "No cluster configurations found in '%s'\n", clusters_dir );
        } else {
            res = format_String( "No cluster configuration '%s' found in '%s'\n", name, clusters_dir );
        }
    }
    destroy( clusters_dir_string );
    return res;
}

/******************************************************************************/
/**                                coord_Init                                **/
/******************************************************************************/
String* coord_Init( const char* conf_dir, const char* cluster, const char* tstr_dir ) //TODO: change to init_MCC
{
    String* res = NULL;

    init_MccData( &MCC );

    if( conf_dir != NULL )
    {
        MCC.config_dir = create_String( conf_dir );
    } else {
        MCC.config_dir = create_String( CONFIG_DIR );
    }
    // TODO: check config dir existence & correctness

    res = set_cluster( cluster );
    if( res != NULL ) return res;

    if( tstr_dir != NULL )
    {
        MCC.testruns_dir = create_String( tstr_dir );
    } else {
        MCC.testruns_dir = create_String( TESTRUNS_DIR );
    }
    // TODO: check testrun dir existence & correctness


    if( start_GetInfoThread() ) {
        res = create_String(GETINFO_THREAD_INIT_ERROR);
    }

    return res;
}

/******************************************************************************/
/**                                runScript                                 **/
/******************************************************************************/
String *getPerrorString( const char *s )
{
    StringBuffer *buf = create_StringBuffer();

    append_StringBuffer( r(buf), s );
    append_StringBuffer( r(buf), ": " );
    append_StringBuffer( r(buf), strerror( errno ) );

    return toString_Object(buf);
}

/******************************************************************************/
static String *runScript( String *scriptName, String *scriptParams )
{
    pid_t pid, cpid = 0;
    int status;
    int pipefd[2], pipefail;
    int stderr_pipefd[2], stderr_pipefail;
    char last_line[LINE_MAX] = "";

    assertion(scriptName != NULL, "runScript: scriptName == NULL");

#ifdef DEBUG_MCC
    printf("MCC: runScript:%s %s\n",
            toCharArray_String( r(scriptName) ), toCharArray_String( r(scriptParams) ));
    fflush(stdout);
#endif
    pipefail = pipe(pipefd);

    // create pipe for stderr
    stderr_pipefail = pipe(stderr_pipefd);

    pid = fork();
    if( pid == -1 ) return getPerrorString( FORK_RUNSCRIPT_ERROR );

    if( pid ) {
        if( !pipefail ) close( pipefd[1] );

        // read stderr and copy the last line 
        if( !stderr_pipefail ) {
            FILE *stderr_file = NULL;
            int line_number = 0;
            int flags;

            // close unused write end of pipe
            close( stderr_pipefd[1] );

            // set nonblocking for reading from pipe
            flags = fcntl(stderr_pipefd[0], F_GETFL);
            fcntl(stderr_pipefd[0], F_SETFL, flags | O_NONBLOCK );

            //read from pipe
            stderr_file = fdopen( stderr_pipefd[0], "r" );
            if( stderr_file != NULL ) {
                while( cpid = waitpid( pid, &status, WNOHANG ) == 0 )
                {
                    char tmp_string[LINE_MAX];
                    char *res = fgets( tmp_string, LINE_MAX, stderr_file );
                    if( res != NULL ) 
                    {
                        strcpy(last_line, tmp_string);
                        //printf("run: stderr: line %d: %s\n", line_number++, tmp_string);
                        fprintf(stderr, tmp_string);
                    }
                }
                fclose(stderr_file);
            } else {
                cpid = waitpid( pid, &status, 0 );
            }
        }

        if( cpid == -1 ) {
            return getPerrorString( WAITPID_RUNSCRIPT_ERROR );
        }
        if( WIFEXITED(status) ) {
            if( WEXITSTATUS(status) ) {
                /* get the error string from child using pipe */
                if( WEXITSTATUS(status) == NOEXEC_CODE ) {
                    if( !pipefail ) {
                        const ssize_t buffsize = 255;
                        char buf[buffsize];
                        ssize_t nbytes;

                        memset( buf, 0, buffsize );
                        nbytes = read( pipefd[0], buf, buffsize );
                        close( pipefd[0] );

                        if( nbytes ) return create_String( buf );
                    }
                }
                {
                    StringBuffer *message = create_StringBuffer();

                    appendFormat_StringBuffer( r(message), NOTZERO_RUNSCRIPT_ERROR " (%d)", WEXITSTATUS(status) );
                    appendFormat_StringBuffer( r(message), ": %s", last_line );
                    return toString_Object( message );
                }
            }
        } else {
            return create_String( NOTNORMALLY_RUNSCRIPT_ERROR );
        }
    } else {
        if( !pipefail ) close( pipefd[0] );

        // redirect stderr to pipe
        if( !stderr_pipefail ) {
            // close unused read end of pipe
            close( stderr_pipefd[0] );
            // redirect stderr to pipe
            dup2( stderr_pipefd[1], 2 );
            close( stderr_pipefd[1] );
        }
        execl( toCharArray_String( r(scriptName) ), toCharArray_String( r(scriptName) )
             , scriptParams == NULL ? NULL : toCharArray_String( r(scriptParams) )
             , NULL );
        /* if exec fails, send the error string to parent using pipe */
        if( !pipefail ) {
            String *message = getPerrorString( EXEC_RUNSCRIPT_ERROR );
            const char *ch = toCharArray_String( r(message) );
            int len = strlen( ch ) + 1;

            write( pipefd[1], ch, len );
            close( pipefd[1] );
        }
        perror( EXEC_RUNSCRIPT_ERROR );
        exit( NOEXEC_CODE );
     }

    return NULL;
}

/******************************************************************************/
/**                               getTestRunId                               **/
/******************************************************************************/
static String *getTestRunId( String *testName )
{
    StringBuffer *buf = create_StringBuffer();
    time_t now;
    struct tm *tm;
    static callNumber = 0;

    assertion(testName != NULL, "getTestRunId: testName == NULL");

    appendString_StringBuffer( r(buf), r(testName) );

    // add the time mark
    now = time( NULL );
    tm = localtime( &now );
    if( tm != NULL ) {
        appendFormat_StringBuffer( r(buf), "_%04d%02d%02d_%02d%02d%02d"
                , tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday
                , tm->tm_hour, tm->tm_min, tm->tm_sec
                );
    }
    // add the call number
    appendFormat_StringBuffer( r(buf), "_%03d", callNumber++);

    return toString_Object( buf );
}

/******************************************************************************/
static String *makeFullName(String *prefix, String *name)
{
    StringBuffer *buf = create_StringBuffer();

    assertion(prefix != NULL, "makeFullName: prefix == NULL");
    assertion(name != NULL, "makeFullName: name == NULL");

    appendString_StringBuffer( r(buf), r(prefix) );
    append_StringBuffer( r(buf), "/" );
    appendString_StringBuffer( r(buf), r(name) );

    return toString_Object( buf );
}

/******************************************************************************/
static String *getStartallScriptName()
{
    String *name = format_String("%s/%s"
                                , coord_getConfigDir()
                                , STARTALL
                                );

    return name;
}

/******************************************************************************/
static String *getEndallScriptName()
{
    String *name = format_String("%s/%s"
                                , coord_getConfigDir()
                                , ENDALL
                                );

    return name;
}

/******************************************************************************/
static String *getKillallScriptName()
{
    String *name = format_String("%s/%s"
                                , coord_getConfigDir()
                                , KILLALL
                                );

    return name;
}



/******************************************************************************/
static String *getNodesFileName(String *testRunId)
{
    String *name = NULL;

    assertion(testRunId != NULL, "getNodesFileName: testRunId == NULL");

    name = format_String("%s/%s/%s"
                        , coord_getTestRunsDir()
                        , toCharArray_String( r(testRunId) )
                        , NODES_OUTFILENAME
                        );

    return name;
}

/******************************************************************************/
static String *getLinksFileName(String *testRunId)
{
    String *name = NULL;

    assertion(testRunId != NULL, "getLinksFileName: testRunId == NULL");

    name = format_String("%s/%s/%s"
                        , coord_getTestRunsDir()
                        , toCharArray_String( r(testRunId) )
                        , LINKS_OUTFILENAME
                        );

    return name;
}

/******************************************************************************/
static String *createParamStringForStart(
          String *testName, String *testParams
        , String *nodes, String *topologyType
        , String *testRunId, String *prevTRI
        )
{
    StringBuffer *buf = NULL;

    assertion(testName != NULL, "createParamStringForStart: testName == NULL");
    assertion(nodes != NULL, "createParamStringForStart: nodes == NULL");
    assertion(topologyType != NULL, "createParamStringForStart: topologyType == NULL");
    assertion(testRunId != NULL, "createParamStringForStart: testRunId == NULL");
    buf = create_StringBuffer();

    append_StringBuffer( r(buf), " " );
    append_StringBuffer( r(buf), coord_getCluster() );

    append_StringBuffer( r(buf), " " );
    appendString_StringBuffer( r(buf), r(nodes) );


    append_StringBuffer( r(buf), " " );
    append_StringBuffer( r(buf), coord_getTestRunsDir() );

    //append_StringBuffer( r(buf), " -i " );
    append_StringBuffer( r(buf), " " );
    appendString_StringBuffer( r(buf), r(testRunId) );

    append_StringBuffer( r(buf), " " );
    appendString_StringBuffer( r(buf), r(topologyType) );

    //append_StringBuffer( r(buf), " -o " );
    append_StringBuffer( r(buf), " " );
    appendString_StringBuffer( r(buf), r(testName) );

    append_StringBuffer( r(buf), " " );
    if( prevTRI == NULL ) {
        append_StringBuffer( r(buf), " \"\"" );
    } else {
        appendString_StringBuffer( r(buf), r(prevTRI) );
    }

    append_StringBuffer( r(buf), " " );
    appendString_StringBuffer( r(buf)
                             , testParams != NULL ? r(testParams) : create_String("\"\"")
                             );

    return toString_Object( buf );
}

/******************************************************************************/
/**                             coord_StartTest                              **/
/******************************************************************************/
String *coord_StartTest(String *testName, String *testParams
        , String *nodes, String *topologyType
        , String *prevTRI, bool saveCP, String **testRunId
)
{
    String *res = NULL;
    String *starter = NULL, *params = NULL;
    String *nodes_cfg = NULL, *links_cfg = NULL;
    TestRun *new_testRun;

    assertion(testRunId != NULL, "coord_StartTest: testRunId == NULL");

    *testRunId = r(getTestRunId( testName ));

    new_testRun = (TestRun *)append_ArrayList( &MCC.testRunList );
    init_TestRun(new_testRun, *testRunId, testName, testParams
               , nodes, topologyType, saveCP, prevTRI);

    /* run startall.sh */
    starter = getStartallScriptName();
    params = createParamStringForStart( testName, testParams, nodes, topologyType, *testRunId, prevTRI );
#ifdef DEBUG_MCC
    printf("MCC: coord_StartTest: runScript\n"); fflush(stdout);
#endif
    res = runScript( starter, params );
#ifdef DEBUG_MCC
    printf("MCC: coord_StartTest: runScript done\n"); fflush(stdout);
#endif
    destroy( starter );
    destroy( params );

    if( res != NULL ) goto ERROR;

    nodes_cfg = getNodesFileName( *testRunId );
    res = setNodes_TestRun( new_testRun, toCharArray_String( r(nodes_cfg) ) );
    destroy( nodes_cfg );
    if( res != NULL ) goto ERROR;

    links_cfg = getLinksFileName( *testRunId );
    res = setLinks_TestRun( new_testRun, toCharArray_String( r(links_cfg) ) );
    destroy( links_cfg );
    if( res != NULL ) goto ERROR;

    res = configureNodes_TestRun( new_testRun );
    if( res != NULL ) goto ERROR;
    new_testRun->status = TESTRUN_CONFIGURED;

    res = startTest_MCC( *testRunId );
    if( res != NULL ) goto ERROR;
    new_testRun->status = TESTRUN_STARTED;

    return NULL;

ERROR:
	new_testRun->status = TESTRUN_FAILED;
	return res;
}

/******************************************************************************/
/**                               KillTest_MCC                               **/
/******************************************************************************/
static String *createParamStringForKill(String *testRunId)
{
    StringBuffer *buf = NULL;

    assertion(testRunId != NULL, "createParamStringForKill: testRunId == NULL");

    buf = create_StringBuffer();

    append_StringBuffer( r(buf), " " );
    append_StringBuffer( r(buf), coord_getCluster() );


    append_StringBuffer( r(buf), " " );
    append_StringBuffer( r(buf), coord_getTestRunsDir() );

    //append_StringBuffer( r(buf), " -i " );
    append_StringBuffer( r(buf), " " );
    appendString_StringBuffer( r(buf), r(testRunId) );

    return toString_Object( buf );
}

/******************************************************************************/
String *KillTest_MCC( String *testRunId )
{
    String *res = NULL;
    String *killer = NULL, *params = NULL;

    assertion(testRunId != NULL, "KillTest_MCC: testRunId == NULL");

    /* run killall.sh */
    killer = getKillallScriptName();
    params = createParamStringForKill(testRunId);
    res = runScript( killer, params );
    destroy( killer );
    destroy( params );

    if( res != NULL ) goto ERROR;

    // TODO: set TestRun to the state KILLED
    return NULL;

ERROR:
    // TODO: set TestRun to the state NOT KILLED
    //new_testRun->status = TESTRUN_FAILED;
    return res;
}
